%% prepare data

% 0 .. nonterminal, 1 .. terminal, 2 .. obstacle
world = [
    0 0 0 1;
    0 2 0 1;
    0 0 0 0
];

actions = [ % w d s a
    -1 0;
    0 1;
    1 0;
    0 -1
];

rewards = -0.04 * ones(size(world));
rewards(1,4) = 1;
rewards(2,4) = -1;

initial_pos = [3 1];

passive_policy = [
    2, 2, 2, 0;
    1, 0, 1, 0;
    1, 4, 4, 4
];

passive_policy2 = [
    2, 2, 2, 0;
    1, 0, 1, 0;
    2, 2, 1, 4
];

save('rl_data.mat', 'world', 'actions', 'rewards', 'initial_pos', 'passive_policy');

%% Data description

% States:
% 0 .. free
% 1 .. final
% 2 .. obstacle

% Actions: 
% 1 2 3 4
% w d s a


%% Config
% 0: no plots; 1: plot after episode; 2: plot every rollout, 3: pause after rollout
plot_level = 3;

%% Passive agent - direct policy estimation

values = % TODO complete statement
values_counts = 1e-6 * ones(size(world));
if (plot_level > 0), figure(1); end;

N = 5000;
for i=1:N
    pos = initial_pos;
    history = []; % x, y, reward
    while (world(pos(1), pos(2)) ~= 1)
        if (plot_level > 1), plot_world(pos, world, values ./ values_counts); end
        
        old_pos = pos;
        action = passive_policy(pos(1), pos(2));
        pos = transition_model(pos, action, world, actions);
        
        % TODO record stuff to history
        if (plot_level > 2), pause; end
    end

    for j=1:size(history,1)
        p = history(j,1:2);
        values(p(1), p(2)) = % TODO complete statement
        values_counts(p(1), p(2)) = values_counts(p(1), p(2)) + 1;
    end
    
    if (plot_level > 0 && mod(i,100) == 1)
        plot_world(pos, world, values ./ values_counts);
    end
    
end
values ./ values_counts

%% Passive agent - Temporal difference

values = zeros(size(world));
alpha = 0.5;
if (plot_level > 0), figure(2); end

N = 20000;
for i=1:N
    pos = initial_pos;
    while (world(pos(1), pos(2)) ~= 1)
        if (plot_level > 1), plot_world(pos, world, values); end
        
        old_pos = pos;
        action = passive_policy(pos(1), pos(2));
        pos = transition_model(pos, action, world, actions);
        
        values(old_pos(1), old_pos(2)) = ...
            values(old_pos(1), old_pos(2)) +...  % U(s)
            alpha *...  % learning rate
            ((N-i)/N) *...  % cooling down the learning rate
            % TODO complete statement
        
        if (plot_level > 2), pause; end
    end
    values(pos(1), pos(2)) = rewards(pos(1), pos(2));
    
    if (plot_level > 0 && mod(i,100) == 1)
        plot_world(pos, world, values);
    end
end
values

%% Active agent - Q-learning

Q = zeros(size(world,1), size(world,2), size(actions,1));
alpha = 0.5;
if (plot_level > 0), figure(3); end;

N = 10000;
for i=1:N
    pos = initial_pos;
    while (world(pos(1), pos(2)) ~= 1)
        if (plot_level > 1), plot_world(pos, world, values); end
        
        old_pos = pos;
        % TODO get action from Q function
        if (rand() > 0.8), action = mod(round(normrnd(action-1, 1.0)), size(actions, 1))+1; end % some exploration
        pos = transition_model(pos, action, world, actions);
        
        Q(old_pos(1), old_pos(2), action) = ...
            Q(old_pos(1), old_pos(2), action) +...  % Q(s,a)
            alpha *...  % learning rate
            ((N-i)/N) *...  % cooling down the learning rate
            % TODO complete statement
        
        if (plot_level > 2), pause; end;
    end
    Q(pos(1), pos(2),:) = rewards(pos(1), pos(2));
    
    if (plot_level > 0 && mod(i,100) == 1)
        plot_world(pos, world, max(Q,[],3));
    end
end
[~,policy] = max(Q,[],3) % show the resulting policy
